#!/usr/bin/env python3

"""Convert the Celestia stars.dat file into an equivalent CSV"""

import struct
import sys
from typing import BinaryIO, TextIO

def decode_id(id: int) -> str:
    """Convert a numeric ID into a HIP or TYC number"""
    if id < 1000000000:
        return f"HIP {id}"
    else:
        tyc1 = id % 10000
        tyc2 = (id // 10000) % 100000
        tyc3 = id // 1000000000
        return f"TYC {tyc1}-{tyc2}-{tyc3}"

NORM_TYPES = ["O","B","A","F","G","K","M","R","S","N","WC","WN","?","L","T","C"]
LUM_TYPES = ["Ia-0","Ia","Ib","II","III","IV","V","VI"]
WD_TYPES = ["A","B","C","O","Q","Z","","X"]

def decode_spec(spec: int) -> str:
    """Convert a numeric spectral type code into a string"""
    k = spec >> 12
    t = (spec >> 8) & 0xf
    s = (spec >> 4) & 0xf
    l = spec & 0xf
    if k == 0:
        # Normal star
        result = NORM_TYPES[t]
        if result == "?":
            return result
        if s < 10:
            result += str(s)
        if l < len(LUM_TYPES):
            result += LUM_TYPES[l]
        return result
    elif k == 1:
        # white dwarf
        result = "D"
        if t < len(WD_TYPES):
            result += WD_TYPES[t]
        if s < 10:
            result += str(s)
        return result
    elif k == 2:
        # neutron star
        return "Q"
    else:
        # black hole
        return "X"

def process_file(infile: BinaryIO, outfile: TextIO) -> bool:
    """Process the stars.dat file"""
    header = infile.read(14)
    magic, version, num_stars = struct.unpack('<8sHI', header)
    # check if this is a stars.dat file
    if magic != b"CELSTARS" or version != 0x0100:
        print("Not a valid stars.dat file", file=sys.stderr)
        return False
    print("id,catName,x,y,z,absMag,spType", file=outfile)
    for i in range(num_stars):
        data = infile.read(20)
        id, x, y, z, absMag, spec = struct.unpack('<I3fhH', data)
        name = decode_id(id)
        absMag /= 256
        sp_type = decode_spec(spec)
        print(f"{id},\"{name}\",{x},{y},{z},{absMag},\"{sp_type}\"", file=outfile)
    return True

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Need to supply path to stars.dat", file=sys.stderr)
        exit(1)
    in_path = sys.argv[1]
    with open(in_path, 'rb') as f_input:
        if len(sys.argv) < 3:
            # second file not specified, use stdout
            ok = process_file(f_input, sys.stdout)
        else:
            out_path = sys.argv[2]
            with open(out_path, 'w') as f_output:
                ok = process_file(f_input, f_output)
    if not ok:
        print("Could not parse input file", file=sys.stderr)
        exit(1)
